"""Test the disease querying method."""

from datetime import datetime

import pytest
from deepdiff import DeepDiff
from ga4gh.core.models import Extension, MappableConcept

from disease.query import InvalidParameterException, QueryHandler
from disease.schemas import MatchType, SourceName


@pytest.fixture(scope="module")
def query_handler(database):
    """Build query handler test fixture"""
    return QueryHandler(database)


@pytest.fixture(scope="module")
def neuroblastoma():
    """Create neuroblastoma fixture."""
    return MappableConcept(
        conceptType="Disease",
        id="normalize.disease.ncit:C3270",
        primaryCoding={
            "id": "ncit:C3270",
            "code": "C3270",
            "system": "https://ncit.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code=",
        },
        name="Neuroblastoma",
        mappings=[
            {
                "coding": {
                    "id": "MONDO_0005072",
                    "code": "MONDO:0005072",
                    "system": "https://purl.obolibrary.org/obo/",
                },
                "relation": "exactMatch",
            },
            {
                "coding": {
                    "id": "oncotree:NBL",
                    "code": "NBL",
                    "system": "https://oncotree.mskcc.org/?version=oncotree_latest_stable&field=CODE&search=",
                },
                "relation": "exactMatch",
            },
            {
                "coding": {
                    "id": "DOID:769",
                    "code": "DOID:769",
                    "system": "https://disease-ontology.org/?id=",
                },
                "relation": "exactMatch",
            },
            {
                "coding": {
                    "id": "umls:C0027819",
                    "code": "C0027819",
                    "system": "https://www.nlm.nih.gov/research/umls/index.html",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "icdo:9500/3",
                    "code": "9500/3",
                    "system": "https://www.who.int/standards/classifications/other-classifications/international-classification-of-diseases-for-oncology/",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "efo:0000621",
                    "code": "0000621",
                    "system": "http://www.ebi.ac.uk/efo/EFO_",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "gard:7185",
                    "code": "7185",
                    "system": "https://rarediseases.info.nih.gov",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "mesh:D009447",
                    "code": "D009447",
                    "system": "https://meshb.nlm.nih.gov/record/ui?ui=",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "orphanet:635",
                    "code": "635",
                    "system": "https://www.orpha.net",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "umls:C2751421",
                    "code": "C2751421",
                    "system": "https://www.nlm.nih.gov/research/umls/index.html",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "medgen:18012",
                    "code": "18012",
                    "system": "https://www.ncbi.nlm.nih.gov/medgen/",
                },
                "relation": "relatedMatch",
            },
        ],
        extensions=[
            Extension(name="oncologic_disease", value=True),
            Extension(
                name="aliases",
                value=[
                    "NB",
                    "neuroblastoma",
                    "Neural Crest Tumor, Malignant",
                    "Neuroblastoma (Schwannian Stroma-poor)",
                    "neuroblastoma (Schwannian Stroma-poor)",
                    "Neuroblastoma (Schwannian Stroma-Poor)",
                    "Neuroblastoma, NOS",
                    "NEUROBLASTOMA, MALIGNANT",
                    "Neuroblastoma (NBL)",
                    "neural Crest tumor, malignant",
                    "neuroblastoma, malignant",
                ],
            ),
        ],
    )


@pytest.fixture(scope="module")
def skin_myo():
    """Create a test fixture for skin myopithelioma"""
    return MappableConcept(
        conceptType="Disease",
        id="normalize.disease.ncit:C167370",
        primaryCoding={
            "id": "ncit:C167370",
            "code": "C167370",
            "system": "https://ncit.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code=",
        },
        name="Skin Myoepithelioma",
        extensions=[Extension(name="aliases", value=["Cutaneous Myoepithelioma"])],
    )


@pytest.fixture(scope="module")
def mafd2():
    """Create a test fixture for major affective disorder 2. Query should not
    include a "pediatric_disease" Extension object.
    """
    return MappableConcept(
        conceptType="Disease",
        id="normalize.disease.mondo:0010648",
        primaryCoding={
            "id": "MONDO_0010648",
            "code": "MONDO:0010648",
            "system": "https://purl.obolibrary.org/obo/",
        },
        name="major affective disorder 2",
        mappings=[
            {
                "coding": {
                    "id": "MIM:309200",
                    "code": "309200",
                    "system": "https://omim.org/MIM:",
                },
                "relation": "exactMatch",
            },
            {
                "coding": {
                    "id": "mesh:C564108",
                    "code": "C564108",
                    "system": "https://meshb.nlm.nih.gov/record/ui?ui=",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "medgen:326975",
                    "code": "326975",
                    "system": "https://www.ncbi.nlm.nih.gov/medgen/",
                },
                "relation": "relatedMatch",
            },
            {
                "coding": {
                    "id": "umls:C1839839",
                    "code": "C1839839",
                    "system": "https://www.nlm.nih.gov/research/umls/index.html",
                },
                "relation": "relatedMatch",
            },
        ],
        extensions=[
            Extension(
                name="aliases",
                value=[
                    "MAFD2",
                    "MDI",
                    "MDX",
                    "MANIC-DEPRESSIVE ILLNESS",
                    "BPAD",
                    "MAFD2",
                    "BIPOLAR AFFECTIVE DISORDER",
                    "MANIC-DEPRESSIVE PSYCHOSIS, X-LINKED",
                    "major affective disorder 2, X-linked dominant",
                    "MAJOR affective disorder 2",
                    "bipolar affective disorder",
                    "major affective disorder 2",
                    "manic-depressive illness",
                    "manic-depressive psychosis, X-linked",
                ],
            )
        ],
    )


def compare_disease(actual, fixture):
    """Verify correctness of returned Disease core object against test fixture."""
    actual = actual.disease
    diff = DeepDiff(fixture, actual, ignore_order=True)
    assert diff == {}, fixture.id


def test_query(query_handler):
    """Test that query returns properly-structured response."""
    resp = query_handler.search("Neuroblastoma")
    assert resp.query == "Neuroblastoma"
    matches = resp.source_matches
    assert isinstance(matches, dict)
    assert len(matches) == 5
    ncit = matches["NCIt"]
    assert len(ncit.records) == 1
    ncit_record = ncit.records[0]
    assert ncit_record.label == "Neuroblastoma"


def test_query_specify_query_handlers(query_handler):
    """Test inclusion and exclusion of sources in query."""
    # test full inclusion
    sources = "ncit,mondo,do,oncotree,omim"
    resp = query_handler.search("Neuroblastoma", incl=sources, excl="")
    matches = resp.source_matches
    assert len(matches) == 5
    assert SourceName.NCIT in matches
    assert SourceName.MONDO in matches
    assert SourceName.DO in matches
    assert SourceName.ONCOTREE in matches
    assert SourceName.OMIM in matches

    # test full exclusion
    resp = query_handler.search("Neuroblastoma", excl=sources)
    matches = resp.source_matches
    assert len(matches) == 0

    # test case insensitive
    resp = query_handler.search("Neuroblastoma", excl="NCIt")
    matches = resp.source_matches
    assert SourceName.NCIT not in matches
    resp = query_handler.search("Neuroblastoma", incl="nCiT")
    matches = resp.source_matches
    assert SourceName.NCIT in matches

    # test error on invalid source names
    with pytest.raises(InvalidParameterException):
        resp = query_handler.search("Neuroblastoma", incl="nct")

    # test error for supplying both incl and excl args
    with pytest.raises(InvalidParameterException):
        resp = query_handler.search("Neuroblastoma", incl="mondo", excl="ncit")


def test_normalize_query(query_handler, neuroblastoma, mafd2):
    """Test that normalized endpoint correctly resolves queries, and utilizes
    the VOD schema.
    """
    response = query_handler.normalize("Neuroblastoma")
    assert response.match_type == MatchType.LABEL
    assert response.warnings is None
    compare_disease(response, neuroblastoma)
    assert len(response.source_meta_) == 4
    assert SourceName.NCIT in response.source_meta_
    assert SourceName.DO in response.source_meta_
    assert SourceName.MONDO in response.source_meta_
    assert SourceName.ONCOTREE in response.source_meta_

    response = query_handler.normalize("MAFD2")
    assert response.match_type == MatchType.ALIAS
    assert response.warnings is None
    compare_disease(response, mafd2)
    assert len(response.source_meta_) == 2
    assert SourceName.MONDO in response.source_meta_


def test_normalize_non_mondo(query_handler, skin_myo, neuroblastoma):
    """Test that normalize endpoint returns records not in Mondo groups."""
    response = query_handler.normalize("Skin Myoepithelioma")
    assert response.match_type == MatchType.LABEL
    assert len(response.source_meta_) == 1
    compare_disease(response, skin_myo)
    assert SourceName.NCIT in response.source_meta_

    response = query_handler.normalize("Cutaneous Myoepithelioma")
    assert response.match_type == MatchType.ALIAS
    assert len(response.source_meta_) == 1
    compare_disease(response, skin_myo)
    assert SourceName.NCIT in response.source_meta_

    response = query_handler.normalize("orphanet:635")
    assert response.match_type == MatchType.XREF
    assert len(response.source_meta_) == 4
    compare_disease(response, neuroblastoma)
    assert SourceName.NCIT in response.source_meta_
    assert SourceName.DO in response.source_meta_
    assert SourceName.MONDO in response.source_meta_
    assert SourceName.ONCOTREE in response.source_meta_


def test_service_meta(query_handler):
    """Test service meta info in response."""
    test_query = "glioma"

    response = query_handler.search(test_query)
    service_meta = response.service_meta_
    assert service_meta.name == "disease-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/disease-normalization"

    response = query_handler.normalize(test_query)
    service_meta = response.service_meta_
    assert service_meta.name == "disease-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/disease-normalization"

    response = query_handler.search("this-will-not-normalize")
    service_meta = response.service_meta_
    assert service_meta.name == "disease-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/disease-normalization"

    response = query_handler.normalize("this-will-not-normalize")
    service_meta = response.service_meta_
    assert service_meta.name == "disease-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/disease-normalization"



"""Get OncoTree data."""

import json
import logging

from tqdm import tqdm

from disease.etl.base import Base
from disease.schemas import NamespacePrefix, SourceMeta

_logger = logging.getLogger(__name__)


class OncoTree(Base):
    """Gather and load data from OncoTree."""

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",  # F401
            version=self._version,
            data_url="http://oncotree.mskcc.org/#/home?tab=api",
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        )
        self._database.add_source_metadata(self._src_name, metadata)

    def _traverse_tree(self, disease_node: dict) -> None:
        """Traverse JSON tree and queue diseases for loading where possible.

        :param disease_node: node in tree containing info for individual disease.
        """
        if disease_node["level"] >= 2:
            self._nodes.append(
                {
                    "code": disease_node["code"],
                    "name": disease_node["name"],
                    "externalReferences": disease_node.get("externalReferences", []),
                }
            )
        if disease_node.get("children"):
            for child in disease_node["children"].values():
                self._traverse_tree(child)

    def _add_disease(self, disease_node: dict) -> None:
        """Grab data from disease node and load into DB.

        :param disease_node: individual node taken from OncoTree tree
        """
        disease = {
            "concept_id": f"{NamespacePrefix.ONCOTREE.value}:{disease_node['code']}",
            "label": disease_node["name"],
            "xrefs": [],
            "associated_with": [],
            "oncologic_disease": True,
        }
        refs = disease_node.get("externalReferences", [])
        for prefix, codes in refs.items():
            if prefix == "UMLS":
                normed_prefix = NamespacePrefix.UMLS.value
                for code in codes:
                    normed_id = f"{normed_prefix}:{code}"
                    disease["associated_with"].append(normed_id)
            elif prefix == "NCI":
                normed_prefix = NamespacePrefix.NCIT.value
                for code in codes:
                    normed_id = f"{normed_prefix}:{code}"
                    disease["xrefs"].append(normed_id)
            else:
                _logger.warning("Unrecognized prefix: %s", prefix)
                continue
        self._load_disease(disease)

    def _transform_data(self) -> None:
        """Initiate OncoTree data transformation."""
        with self._data_file.open() as f:
            oncotree = json.load(f)
        self._nodes = []
        self._traverse_tree(oncotree["TISSUE"])
        for node in tqdm(self._nodes, ncols=80, disable=self._silent):
            self._add_disease(node)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Disease Normalizer"
copyright = "2021-2024, Wagner Lab"
author = "Wagner Lab"
html_title = "disease-normalizer"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
    "sphinx.ext.autosummary",
    "sphinx_github_changelog",
    "sphinx_click",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = []
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/fontawesome.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/cancervariants/disease-normalization",
            "html": "",
            "class": "fa-brands fa-solid fa-github",
        },
        {
            "name": "Wagner Lab",
            "url": "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab",
            "html": "",
            "class": "fa-solid fa-house",
        }
    ],
}
# -- autodoc things ----------------------------------------------------------
import os
import sys

sys.path.insert(0, os.path.abspath("../../src/disease"))
autodoc_preserve_defaults = True

# -- get version -------------------------------------------------------------
from disease import __version__

release = version = __version__

# -- linkcode ----------------------------------------------------------------
def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"https://github.com/cancervariants/disease-normalization/blob/main/src/{filename}.py"


# -- code block style --------------------------------------------------------
pygments_style = "default"
pygements_dark_style = "monokai"


# -- sphinx-click ------------------------------------------------------------
# These functions let us write descriptions/docstrings in a way that doesn't look
# weird in the Click CLI, but get additional formatting in the sphinx-click autodocs for
# better readability.
import re

from click.core import Context
from sphinx.application import Sphinx
from sphinx_click.ext import _get_usage, _indent


CMD_PATTERN = r"--[^ ]+"
STR_PATTERN = r"\"[^ ]+\""
SNAKE_PATTERN = r"[A-Z]+_[A-Z_]*[A-Z][., ]"


def _add_formatting_to_string(line: str) -> str:
    """Add fixed-width code formatting to span sections in lines:

    * shell options, eg `--update_all`
    * double-quoted strings, eg `"DO"`
    * all caps SNAKE_CASE env vars, eg `DISEASE_NORM_REMOTE_DB_URL`
    """
    for pattern in (CMD_PATTERN, STR_PATTERN, SNAKE_PATTERN):
        line = re.sub(pattern, lambda x: f"``{x.group()}``", line)
    return line


def process_description(app: Sphinx, ctx: Context, lines: list[str]):
    """Add custom formatting to sphinx-click autodoc descriptions.

    * remove :param: :return: etc
    * add fixed-width (code) font to certain words
    * add code block formatting to example shell commands
    * move primary usage example to the top of the description

    Because we have to modify the lines list in place, we have to make multiple passes
    through it to format everything correctly.
    """
    if not lines:
        return

    # chop off params
    param_boundary = None
    for i, line in enumerate(lines):
        if ":param" in line:
            param_boundary = i
            break
    if param_boundary is not None:
        del lines[param_boundary:]
        lines[-1] = ""

    # add code formatting to strings, commands, and env vars
    lines_to_fmt = []
    for i, line in enumerate(lines):
        if line.startswith(("   ", ">>> ", "|")):
            continue  # skip example code blocks
        if any(
            [
                re.findall(CMD_PATTERN, line),
                re.findall(STR_PATTERN, line),
                re.findall(SNAKE_PATTERN, line),
            ]
        ):
            lines_to_fmt.append(i)
    for line_num in lines_to_fmt:
        lines[line_num] = _add_formatting_to_string(lines[line_num])

    # add code block formatting to example console commands
    for i in range(len(lines) - 1, -1, -1):
        if lines[i].startswith(("    ", "|     ")):
            if lines[i].startswith("|     "):
                lines[i] = lines[i][3:]
            if (i == 0 or lines[i - 1] == "\b" or lines[i - 1] == ""):
                lines.insert(i, "")
                lines.insert(i, ".. code-block:: console")

    # put usage at the top of the description
    lines.insert(0, "")
    for usage_line in _get_usage(ctx).splitlines()[::-1]:
        lines.insert(0, _indent(usage_line))
    lines.insert(0, "")
    lines.insert(0, ".. code-block:: shell")


def process_option(app: Sphinx, ctx: Context, lines: list[str]):
    """Add fixed-width formatting to strings in sphinx-click autodoc options."""
    for i, line in enumerate(lines):
        if re.findall(STR_PATTERN, line):
            lines[i] = re.sub(STR_PATTERN, lambda x: f"``{x.group()}``", line)


def setup(app):
    """Used to hook format customization into sphinx-click build.

    In particular, since we move usage to the top of the command description, we need
    an extra hook here to silence the built-in usage section.
    """
    app.connect("sphinx-click-process-description", process_description)
    app.connect("sphinx-click-process-options", process_option)
    app.connect("sphinx-click-process-usage", lambda app, ctx, lines: lines.clear())



"""The VICC library for normalizing diseases."""

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

APP_ROOT = Path(__file__).resolve().parents[0]


try:
    __version__ = version("disease-normalizer")
except PackageNotFoundError:
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from disease.schemas import (  # noqa: E402
    NamespacePrefix,
    RefType,
    SourceIDAfterNamespace,
    SourceName,
)

ITEM_TYPES = {k.lower(): v.value for k, v in RefType.__members__.items()}

# use to lookup source name from lower-case string
# technically the same as PREFIX_LOOKUP, but source namespace prefixes
# sometimes differ from their names
# e.g. {'ncit': 'NCIt'}
SOURCES_LOWER_LOOKUP = {
    name.value.lower(): name.value for name in SourceName.__members__.values()
}

# use to fetch source name from schema based on concept id namespace
# e.g. {'ncit': 'NCIt', 'doid': 'DO'}
PREFIX_LOOKUP = {
    v.value.lower(): SourceName[k].value
    for k, v in NamespacePrefix.__members__.items()
    if k in SourceName.__members__
}

# use to generate namespace prefix from source ID value
# e.g. {'c': 'NCIt'}
NAMESPACE_LOOKUP = {
    v.value.lower(): NamespacePrefix[k].value
    for k, v in SourceIDAfterNamespace.__members__.items()
    if v.value != ""
}

# Use for checking whether to pull IDs for merge group generation
SOURCES_FOR_MERGE = {SourceName.MONDO}



"""Provide DynamoDB client."""

import atexit
import logging
import sys
from collections.abc import Generator
from os import environ
from pathlib import Path
from typing import Any

import boto3
import click
from boto3.dynamodb.conditions import Attr, Equals, Key
from botocore.exceptions import ClientError

from disease import ITEM_TYPES, PREFIX_LOOKUP
from disease.database.database import (
    AWS_ENV_VAR_NAME,
    SKIP_AWS_DB_ENV_NAME,
    VALID_AWS_ENV_NAMES,
    AbstractDatabase,
    AwsEnvName,
    DatabaseException,
    DatabaseReadException,
    DatabaseWriteException,
    confirm_aws_db_use,
)
from disease.schemas import (
    DataLicenseAttributes,
    RecordType,
    RefType,
    SourceMeta,
    SourceName,
)

_logger = logging.getLogger()


class DynamoDbDatabase(AbstractDatabase):
    """Disease Normalizer database client for DynamoDB."""

    def __init__(self, db_url: str | None = None, **db_args) -> None:
        """Initialize Database class.

        :param str db_url: URL endpoint for DynamoDB source
        :Keyword Arguments:
            * region_name: AWS region (defaults to "us-east-2")
        """
        self.disease_table = environ.get("DISEASE_DYNAMO_TABLE", "disease_normalizer")

        region_name = db_args.get("region_name", "us-east-2")

        if AWS_ENV_VAR_NAME in environ:
            if "DISEASE_NORM_TEST" in environ:
                msg = f"Cannot have both DISEASE_NORM_TEST and {AWS_ENV_VAR_NAME} set."
                raise DatabaseException(msg)

            aws_env = environ[AWS_ENV_VAR_NAME]
            if aws_env not in VALID_AWS_ENV_NAMES:
                msg = f"{AWS_ENV_VAR_NAME} must be one of {VALID_AWS_ENV_NAMES}"
                raise DatabaseException(msg)

            skip_confirmation = environ.get(SKIP_AWS_DB_ENV_NAME)
            if (not skip_confirmation) or (
                skip_confirmation and skip_confirmation != "true"
            ):
                confirm_aws_db_use(environ[AWS_ENV_VAR_NAME])

            boto_params = {"region_name": region_name}

            if aws_env == AwsEnvName.DEVELOPMENT:
                self.disease_table = environ.get(
                    "DISEASE_DYNAMO_TABLE", "disease_normalizer_nonprod"
                )
        else:
            if db_url:
                endpoint_url = db_url
            elif "DISEASE_NORM_DB_URL" in environ:
                endpoint_url = environ["DISEASE_NORM_DB_URL"]
            else:
                endpoint_url = "http://localhost:8000"
            click.echo(f"***Using Disease Database Endpoint: {endpoint_url}***")
            boto_params = {"region_name": region_name, "endpoint_url": endpoint_url}

        self.dynamodb = boto3.resource("dynamodb", **boto_params)
        self.dynamodb_client = boto3.client("dynamodb", **boto_params)

        # Only create tables for local instance
        envs_do_not_create_tables = {AWS_ENV_VAR_NAME, "DISEASE_NORM_TEST"}
        if not set(envs_do_not_create_tables) & set(environ):
            self.initialize_db()

        self.diseases = self.dynamodb.Table(self.disease_table)
        self.batch = self.diseases.batch_writer()
        self._cached_sources: dict[str, SourceMeta] = {}
        atexit.register(self.close_connection)

    def list_tables(self) -> list[str]:
        """Return names of tables in database.

        :return: Table names in DynamoDB
        """
        return self.dynamodb_client.list_tables()["TableNames"]

    def drop_db(self) -> None:
        """Delete all tables from database. Requires manual confirmation.

        :raise DatabaseWriteException: if called in a protected setting with
            confirmation silenced.
        """
        try:
            if not self._check_delete_okay():
                return
        except DatabaseWriteException:
            _logger.exception("Encountered DB write error when trying to drop DB")
            raise
        if self.disease_table in self.list_tables():
            self.dynamodb.Table(self.disease_table).delete()

    def _create_diseases_table(self) -> None:
        """Create Diseases table."""
        self.dynamodb.create_table(
            TableName=self.disease_table,
            KeySchema=[
                {"AttributeName": "label_and_type", "KeyType": "HASH"},  # Partition key
                {"AttributeName": "concept_id", "KeyType": "RANGE"},  # Sort key
            ],
            AttributeDefinitions=[
                {"AttributeName": "label_and_type", "AttributeType": "S"},
                {"AttributeName": "concept_id", "AttributeType": "S"},
                {"AttributeName": "src_name", "AttributeType": "S"},
                {"AttributeName": "item_type", "AttributeType": "S"},
            ],
            GlobalSecondaryIndexes=[
                {
                    "IndexName": "src_index",
                    "KeySchema": [{"AttributeName": "src_name", "KeyType": "HASH"}],
                    "Projection": {"ProjectionType": "KEYS_ONLY"},
                    "ProvisionedThroughput": {
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                },
                {
                    "IndexName": "item_type_index",
                    "KeySchema": [{"AttributeName": "item_type", "KeyType": "HASH"}],
                    "Projection": {"ProjectionType": "KEYS_ONLY"},
                    "ProvisionedThroughput": {
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                },
            ],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

    def check_schema_initialized(self) -> bool:
        """Check if database schema is properly initialized.

        :return: True if DB appears to be fully initialized, False otherwise
        """
        existing_tables = self.list_tables()
        exists = self.disease_table in existing_tables
        if not exists:
            _logger.info("%s is missing or unavailable.", self.disease_table)
        return exists

    def check_tables_populated(self) -> bool:
        """Perform rudimentary checks to see if tables are populated.

        Emphasis is on rudimentary -- if some fiendish element has deleted half of the
        disease aliases, this method won't pick it up. It just wants to see if a few
        critical tables have at least a small number of records.

        :return: True if queries successful, false if DB appears empty
        """
        sources = self.diseases.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq("source"),
        ).get("Items", [])
        if len(sources) < len(SourceName):
            _logger.info("Disease sources table is missing expected sources.")
            return False

        records = self.diseases.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq(RecordType.IDENTITY.value),
            Limit=1,
        )
        if len(records.get("Items", [])) < 1:
            _logger.info("Disease records index is empty.")
            return False

        normalized_records = self.diseases.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq(RecordType.MERGER.value),
            Limit=1,
        )
        if len(normalized_records.get("Items", [])) < 1:
            _logger.info("Normalized disease records index is empty.")
            return False

        return True

    def initialize_db(self) -> None:
        """Create disease_normalizer table if needed."""
        if not self.check_schema_initialized():
            self._create_diseases_table()

    def get_source_metadata(self, src_name: str | SourceName) -> SourceMeta | None:
        """Get license, versioning, data lookup, etc information for a source.

        :param src_name: name of the source to get data for
        :return: source metadata, if lookup is successful
        """
        if isinstance(src_name, SourceName):
            src_name = src_name.value
        if src_name in self._cached_sources:
            return self._cached_sources[src_name]
        pk = f"{src_name.lower()}##source"
        concept_id = f"source:{src_name.lower()}"
        retrieved_metadata = self.diseases.get_item(
            Key={"label_and_type": pk, "concept_id": concept_id}
        ).get("Item")
        if not retrieved_metadata:
            return None
        formatted_metadata = SourceMeta(
            data_license=retrieved_metadata["data_license"],
            data_license_url=retrieved_metadata["data_license_url"],
            version=retrieved_metadata["version"],
            data_url=retrieved_metadata["data_url"],
            rdp_url=retrieved_metadata["rdp_url"],
            data_license_attributes=DataLicenseAttributes(
                **retrieved_metadata["data_license_attributes"]
            ),
        )
        self._cached_sources[src_name] = formatted_metadata
        return formatted_metadata

    def get_record_by_id(
        self, concept_id: str, case_sensitive: bool = True, merge: bool = False
    ) -> dict | None:
        """Fetch record corresponding to provided concept ID

        :param str concept_id: concept ID for disease record
        :param bool case_sensitive: if true, performs exact lookup, which is more
            efficient. Otherwise, performs filter operation, which doesn't require
            correct casing.
        :param bool merge: if true, look for merged record; look for identity record
            otherwise.
        :return: complete record, if match is found; None otherwise
        """
        try:
            if merge:
                pk = f"{concept_id.lower()}##{RecordType.MERGER.value}"
            else:
                pk = f"{concept_id.lower()}##{RecordType.IDENTITY.value}"
            if case_sensitive:
                response = self.diseases.get_item(
                    Key={"label_and_type": pk, "concept_id": concept_id}
                )
                record = response.get("Item")
                if not record:
                    return None
            else:
                exp = Key("label_and_type").eq(pk)
                response = self.diseases.query(KeyConditionExpression=exp)
                record = response["Items"][0]
            del record["label_and_type"]
        except ClientError as e:
            _logger.exception(
                "boto3 client error on get_records_by_id for search term %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )
            return None
        except (KeyError, IndexError):  # record doesn't exist
            return None
        else:
            return record

    def get_refs_by_type(self, search_term: str, ref_type: RefType) -> list[str]:
        """Retrieve concept IDs for records matching the user's query. Other methods
        are responsible for actually retrieving full records.

        :param search_term: string to match against
        :param ref_type: type of match to look for.
        :return: list of associated concept IDs. Empty if lookup fails.
        """
        pk = f"{search_term}##{ref_type.value.lower()}"
        filter_exp = Key("label_and_type").eq(pk)
        try:
            matches = self.diseases.query(KeyConditionExpression=filter_exp)
            return [m["concept_id"] for m in matches.get("Items", None)]
        except ClientError as e:
            _logger.exception(
                "boto3 client error on get_refs_by_type for search term %s: %s",
                search_term,
                e.response["Error"]["Message"],
            )
            return []

    def get_all_concept_ids(self, source: SourceName | None = None) -> set[str]:
        """Retrieve concept IDs for use in generating normalized records.

        :param source: optionally, just get all IDs for a specific source
        :return: Set of concept IDs as strings.
        """
        last_evaluated_key = None
        concept_ids = []
        params: dict[str, str | Equals] = {
            "ProjectionExpression": "concept_id",
        }
        if source is not None:
            params["FilterExpression"] = Attr("src_name").eq(source.value)
        while True:
            if last_evaluated_key:
                response = self.diseases.scan(
                    ExclusiveStartKey=last_evaluated_key, **params
                )
            else:
                response = self.diseases.scan(**params)
            records = response["Items"]
            concept_ids.extend(record["concept_id"] for record in records)
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                break
        return set(concept_ids)

    def get_all_records(self, record_type: RecordType) -> Generator[dict, None, None]:
        """Retrieve all source or normalized records. Either return all source records,
        or all records that qualify as "normalized" (i.e., merged groups + source
        records that are otherwise ungrouped).
        For example,

        .. code-block:: pycon

           >>> from disease.database import create_db
           >>> from disease.schemas import RecordType
           >>> db = create_db()
           >>> for record in db.get_all_records(RecordType.MERGER):
           >>>     pass  # do something

        :param record_type: type of result to return
        :return: Generator that lazily provides records as they are retrieved
        """
        last_evaluated_key = None
        while True:
            if last_evaluated_key:
                response = self.diseases.scan(
                    ExclusiveStartKey=last_evaluated_key,
                )
            else:
                response = self.diseases.scan()
            records = response.get("Items", [])
            for record in records:
                incoming_record_type = record.get("item_type")
                if record_type == RecordType.IDENTITY:
                    if incoming_record_type == record_type:
                        yield record
                else:
                    if (
                        incoming_record_type == RecordType.IDENTITY
                        and not record.get("merge_ref")
                    ) or incoming_record_type == RecordType.MERGER:
                        yield record
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                break

    def add_source_metadata(self, src_name: SourceName, meta: SourceMeta) -> None:
        """Add new source metadata entry.

        :param src_name: name of source
        :param meta: known source attributes
        :raise DatabaseWriteException: if write fails
        """
        src_name_value = src_name.value
        metadata_item = meta.model_dump()
        metadata_item["src_name"] = src_name_value
        metadata_item["label_and_type"] = f"{str(src_name_value).lower()}##source"
        metadata_item["concept_id"] = f"source:{str(src_name_value).lower()}"
        metadata_item["item_type"] = "source"
        try:
            self.diseases.put_item(Item=metadata_item)
        except ClientError as e:
            raise DatabaseWriteException(e) from e

    def add_record(self, record: dict, src_name: SourceName) -> None:
        """Add new record to database.

        :param Dict record: record to upload
        :param SourceName src_name: name of source for record
        """
        concept_id = record["concept_id"]
        record["src_name"] = src_name.value
        label_and_type = f"{concept_id.lower()}##identity"
        record["label_and_type"] = label_and_type
        record["item_type"] = "identity"
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.exception(
                "boto3 client error on add_record for %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )
        for attr_type, item_type in ITEM_TYPES.items():
            if attr_type in record:
                value = record.get(attr_type)
                if not value:
                    continue
                if isinstance(value, str):
                    items = [value.lower()]
                else:
                    items = {item.lower() for item in value}
                for item in items:
                    self._add_ref_record(
                        item, record["concept_id"], item_type, src_name
                    )

    def add_merged_record(self, record: dict) -> None:
        """Add merged record to database.

        :param record: merged record to add
        """
        concept_id = record["concept_id"]
        id_prefix = concept_id.split(":")[0].lower()
        record["src_name"] = PREFIX_LOOKUP[id_prefix]
        label_and_type = f"{concept_id.lower()}##{RecordType.MERGER.value}"
        record["label_and_type"] = label_and_type
        record["item_type"] = RecordType.MERGER.value
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.exception(
                "boto3 client error on add_record for %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )

    def _add_ref_record(
        self, term: str, concept_id: str, ref_type: str, src_name: SourceName
    ) -> None:
        """Add auxiliary/reference record to database.

        :param str term: referent term
        :param str concept_id: concept ID to refer to
        :param str ref_type: one of {'alias', 'label', 'xref',
            'associated_with'}
        :param src_name: name of source for record
        """
        label_and_type = f"{term.lower()}##{ref_type}"
        record = {
            "label_and_type": label_and_type,
            "concept_id": concept_id.lower(),
            "src_name": src_name.value,
            "item_type": ref_type,
        }
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.exception(
                "boto3 client error adding reference %s for %s with match type %s: %s",
                term,
                concept_id,
                ref_type,
                e.response["Error"]["Message"],
            )

    def update_merge_ref(self, concept_id: str, merge_ref: Any) -> None:  # noqa: ANN401
        """Update the merged record reference of an individual record to a new value.

        :param concept_id: record to update
        :param merge_ref: new ref value
        :raise DatabaseWriteException: if attempting to update non-existent record
        """
        label_and_type = f"{concept_id.lower()}##identity"
        key = {"label_and_type": label_and_type, "concept_id": concept_id}
        update_expression = "set merge_ref=:r"
        update_values = {":r": merge_ref.lower()}
        condition_expression = "attribute_exists(label_and_type)"
        try:
            self.diseases.update_item(
                Key=key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=update_values,
                ConditionExpression=condition_expression,
            )
        except ClientError as e:
            code = e.response.get("Error", {}).get("Code")
            if code == "ConditionalCheckFailedException":
                msg = f"No such record exists for keys {label_and_type}, {concept_id}"
                raise DatabaseWriteException(msg) from e
            _logger.exception(
                "boto3 client error in `database.update_record()`: %s",
                e.response["Error"]["Message"],
            )

    def delete_normalized_concepts(self) -> None:
        """Remove merged records from the database. Use when performing a new update
        of normalized data.

        :raise DatabaseReadException: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteException: if deletion call fails
        """
        while True:
            with self.diseases.batch_writer(
                overwrite_by_pkeys=["label_and_type", "concept_id"]
            ) as batch:
                try:
                    response = self.diseases.query(
                        IndexName="item_type_index",
                        KeyConditionExpression=Key("item_type").eq(
                            RecordType.MERGER.value
                        ),
                    )
                except ClientError as e:
                    raise DatabaseReadException(e) from e
                records = response["Items"]
                if not records:
                    break
                for record in records:
                    batch.delete_item(
                        Key={
                            "label_and_type": record["label_and_type"],
                            "concept_id": record["concept_id"],
                        }
                    )

    def delete_source(self, src_name: SourceName) -> None:
        """Delete all data for a source. Use when updating source data.

        :param src_name: name of source to delete
        :raise DatabaseReadException: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteException: if deletion call fails
        """
        while True:
            try:
                response = self.diseases.query(
                    IndexName="src_index",
                    KeyConditionExpression=Key("src_name").eq(src_name.value),
                )
            except ClientError as e:
                raise DatabaseReadException(e) from e
            records = response["Items"]
            if not records:
                break
            with self.diseases.batch_writer(
                overwrite_by_pkeys=["label_and_type", "concept_id"]
            ) as batch:
                try:
                    for record in records:
                        batch.delete_item(
                            Key={
                                "label_and_type": record["label_and_type"],
                                "concept_id": record["concept_id"],
                            }
                        )
                except ClientError as e:
                    raise DatabaseWriteException(e) from e

    def complete_write_transaction(self) -> None:
        """Conclude transaction or batch writing if relevant."""
        self.batch.__exit__(*sys.exc_info())
        self.batch = self.diseases.batch_writer()

    def close_connection(self) -> None:
        """Perform any manual connection closure procedures if necessary."""
        self.batch.__exit__(*sys.exc_info())

    def load_from_remote(self, url: str | None = None) -> None:
        """Load DB from remote dump. Not available for DynamoDB database backend.

        :param url: remote location to retrieve gzipped dump file from
        """
        raise NotImplementedError

    def export_db(self, export_location: Path) -> None:
        """Dump DB to specified location. Not available for DynamoDB database backend.

        :param export_location: path to save DB dump at
        """
        raise NotImplementedError



"""Provide miscellaneous utilities related to disease term normalization."""

import logging
from collections.abc import Generator

from disease.database import AbstractDatabase
from disease.schemas import RecordType, SourceName


def get_term_mappings(
    database: AbstractDatabase,
    scope: RecordType | SourceName,
    cancer_only: bool = False,
) -> Generator[dict, None, None]:
    """Produce dict objects for known concepts (name + ID) plus other possible referents

    Use in downstream applications such as autocompletion.

    :param database: instance of DB connection to get records from
    :param scope: constrain record scope, either to a kind of record or to a specific source
    :param cancer_only: whether to just include cancer terms (not supported by all sources)
    :return: Generator yielding mapping objects
    """
    if isinstance(scope, SourceName):
        record_type = RecordType.IDENTITY
        src_name = scope
    else:
        record_type = scope
        src_name = None

    for record in database.get_all_records(record_type=record_type):
        if src_name and record["src_name"] != src_name:
            continue
        if cancer_only and record.get("oncologic_disease") is not True:
            continue

        yield {
            "concept_id": record["concept_id"],
            "label": record["label"],
            "aliases": record.get("aliases", []),
            "xrefs": record.get("xrefs", []) + record.get("associated_with", []),
        }


def initialize_logs(log_level: int = logging.INFO) -> None:
    """Configure logging.

    :param log_level: app log level to set
    """
    logging.basicConfig(
        filename=f"{__package__}.log",
        format="[%(asctime)s] - %(name)s - %(levelname)s : %(message)s",
    )
    logger = logging.getLogger(__package__)
    logger.setLevel(log_level)



"""Test OMIM source."""

import re

import pytest

from disease.etl.omim import OMIM
from disease.schemas import Disease, MatchType


@pytest.fixture(scope="module")
def omim(test_source):
    """Build OMIM ETL test fixture."""
    return test_source(OMIM)


@pytest.fixture(scope="module")
def mafd2():
    """Build MAFD2 test fixture."""
    return Disease(
        concept_id="MIM:309200",
        label="MAJOR AFFECTIVE DISORDER 2",
        aliases=[
            "MAFD2",
            "MANIC-DEPRESSIVE ILLNESS",
            "MDI",
            "MANIC-DEPRESSIVE PSYCHOSIS, X-LINKED",
            "MDX",
            "BIPOLAR AFFECTIVE DISORDER",
            "BPAD",
        ],
        xrefs=[],
        associated_with=[],
        pediatric_disease=None,
        oncologic_disease=None,
    )


@pytest.fixture(scope="module")
def acute_ll():
    """Build ALL fixture."""
    return Disease(
        concept_id="MIM:613065",
        label="LEUKEMIA, ACUTE LYMPHOBLASTIC",
        aliases=[
            "ALL",
            "LEUKEMIA, ACUTE LYMPHOBLASTIC, SUSCEPTIBILITY TO, 1",
            "ALL1",
            "LEUKEMIA, ACUTE LYMPHOCYTIC, SUSCEPTIBILITY TO, 1",
            "LEUKEMIA, B-CELL ACUTE LYMPHOBLASTIC, SUSCEPTIBILITY TO",
            "LEUKEMIA, T-CELL ACUTE LYMPHOBLASTIC, SUSCEPTIBILITY TO",
            "LEUKEMIA, ACUTE LYMPHOBLASTIC, B-HYPERDIPLOID, SUSCEPTIBILITY TO",
        ],
        xrefs=[],
        associated_with=[],
        pediatric_disease=None,
        oncologic_disease=None,
    )


@pytest.fixture(scope="module")
def lall():
    """Build LALL fixture."""
    return Disease(
        concept_id="MIM:247640",
        label="LYMPHOBLASTIC LEUKEMIA, ACUTE, WITH LYMPHOMATOUS FEATURES",
        aliases=["LALL", "LYMPHOMATOUS ALL"],
        xrefs=[],
        associated_with=[],
        pediatric_disease=None,
        oncologic_disease=None,
    )


def test_concept_id_match(omim, mafd2, acute_ll, lall, compare_response):
    """Test concept ID search resolution."""
    response = omim.search("MIM:309200")
    compare_response(response, MatchType.CONCEPT_ID, mafd2)

    response = omim.search("MIM:613065")
    compare_response(response, MatchType.CONCEPT_ID, acute_ll)

    response = omim.search("MIM:247640")
    compare_response(response, MatchType.CONCEPT_ID, lall)


def test_label_match(omim, mafd2, acute_ll, lall, compare_response):
    """Test label search resolution."""
    response = omim.search("MAJOR AFFECTIVE DISORDER 2")
    compare_response(response, MatchType.LABEL, mafd2)

    response = omim.search("LEUKEMIA, ACUTE LYMPHOBLASTIC")
    compare_response(response, MatchType.LABEL, acute_ll)

    response = omim.search("lymphoblastic leukemia, acute, with lymphomatous features")
    compare_response(response, MatchType.LABEL, lall)


def test_alias_match(omim, mafd2, acute_ll, lall, compare_response):
    """Test alias search resolution."""
    response = omim.search("manic-depressive psychosis, x-linked")
    compare_response(response, MatchType.ALIAS, mafd2)

    response = omim.search("mafd2")
    compare_response(response, MatchType.ALIAS, mafd2)

    response = omim.search("all")
    compare_response(response, MatchType.ALIAS, acute_ll)

    response = omim.search("LEUKEMIA, ACUTE LYMPHOCYTIC, SUSCEPTIBILITY TO, 1")
    compare_response(response, MatchType.ALIAS, acute_ll)

    response = omim.search("LEUKEMIA, B-CELL ACUTE LYMPHOBLASTIC, SUSCEPTIBILITY TO")
    compare_response(response, MatchType.ALIAS, acute_ll)

    response = omim.search("lymphomatous all")
    compare_response(response, MatchType.ALIAS, lall)

    response = omim.search("lall")
    compare_response(response, MatchType.ALIAS, lall)


def test_meta(omim):
    """Test that meta field is correct."""
    response = omim.search("irrelevant-search-string")
    assert response.source_meta_.data_license == "custom"
    assert response.source_meta_.data_license_url == "https://omim.org/help/agreement"
    assert re.match(r"\d{4}\d{2}\d{2}", response.source_meta_.version)
    assert response.source_meta_.data_url == "https://www.omim.org/downloads"
    assert response.source_meta_.rdp_url == "http://reusabledata.org/omim.html"
    assert not response.source_meta_.data_license_attributes.non_commercial
    assert response.source_meta_.data_license_attributes.share_alike
    assert response.source_meta_.data_license_attributes.attribution



"""Test Oncotree ETL methods."""

import re

import pytest

from disease.etl.oncotree import OncoTree
from disease.schemas import Disease, MatchType


@pytest.fixture(scope="module")
def oncotree(test_source):
    """Build OncoTree ETL test fixture."""
    return test_source(OncoTree)


@pytest.fixture(scope="module")
def neuroblastoma():
    """Return neuroblastoma test fixture."""
    return Disease(
        label="Neuroblastoma",
        concept_id="oncotree:NBL",
        aliases=[],
        xrefs=["ncit:C3270"],
        associated_with=["umls:C0027819"],
        pediatric_disease=None,
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def nsclc():
    """Return non small cell lung cancer fixture"""
    return Disease(
        label="Non-Small Cell Lung Cancer",
        concept_id="oncotree:NSCLC",
        aliases=[],
        xrefs=["ncit:C2926"],
        associated_with=["umls:C0007131"],
        pediatric_disease=None,
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def ipn():
    """Return fixture for intracholecystic papillary neoplasm"""
    return Disease(
        label="Intracholecystic Papillary Neoplasm",
        concept_id="oncotree:ICPN",
        aliases=[],
        xrefs=[],
        associated_with=[],
        pediatric_disease=None,
        oncologic_disease=True,
    )


def test_concept_id_match(oncotree, neuroblastoma, nsclc, ipn, compare_response):
    """Test that concept ID search resolves to correct record"""
    response = oncotree.search("oncotree:NBL")
    compare_response(response, MatchType.CONCEPT_ID, neuroblastoma)

    response = oncotree.search("oncotree:NSCLC")
    compare_response(response, MatchType.CONCEPT_ID, nsclc)

    response = oncotree.search("oncotree:icpn")
    compare_response(response, MatchType.CONCEPT_ID, ipn)

    response = oncotree.search("oncotree:ICPN")
    compare_response(response, MatchType.CONCEPT_ID, ipn)

    response = oncotree.search("ipn")
    assert response.match_type == MatchType.NO_MATCH


def test_label_match(oncotree, neuroblastoma, nsclc, ipn, compare_response):
    """Test that label search resolves to correct record."""
    response = oncotree.search("Neuroblastoma")
    compare_response(response, MatchType.LABEL, neuroblastoma)

    response = oncotree.search("NEUROBLASTOMA")
    compare_response(response, MatchType.LABEL, neuroblastoma)

    response = oncotree.search("non-small cell lung cancer")
    compare_response(response, MatchType.LABEL, nsclc)

    response = oncotree.search("intracholecystic papillary neoplasm")
    compare_response(response, MatchType.LABEL, ipn)


def test_associated_with_match(oncotree, neuroblastoma, nsclc, compare_response):
    """Test that associated_with search resolves to correct record."""
    response = oncotree.search("umls:c0027819")
    compare_response(response, MatchType.ASSOCIATED_WITH, neuroblastoma)

    response = oncotree.search("umls:C0007131")
    compare_response(response, MatchType.ASSOCIATED_WITH, nsclc)


def test_meta(oncotree):
    """Test that meta field is correct."""
    response = oncotree.search("neuroblastoma")
    assert response.source_meta_.data_license == "CC BY 4.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by/4.0/legalcode"
    )
    assert re.match(r"\d\d\d\d\d\d\d\d", response.source_meta_.version)
    assert response.source_meta_.data_url == "http://oncotree.mskcc.org/#/home?tab=api"
    assert response.source_meta_.rdp_url is None
    assert not response.source_meta_.data_license_attributes.non_commercial
    assert not response.source_meta_.data_license_attributes.share_alike
    assert response.source_meta_.data_license_attributes.attribution



"""Test NCIt source."""

import re

import pytest

from disease.etl.ncit import NCIt
from disease.schemas import Disease, MatchType


@pytest.fixture(scope="module")
def ncit(test_source):
    """Build NCIt ETL test fixture."""
    return test_source(NCIt)


@pytest.fixture(scope="module")
def neuroblastoma():
    """Build neuroblastoma test fixture."""
    return Disease(
        label_and_type="ncit:c3270##identity",
        concept_id="ncit:C3270",
        label="Neuroblastoma",
        aliases=[
            "Neural Crest Tumor, Malignant",
            "Neuroblastoma (NBL)",
            "Neuroblastoma (Schwannian Stroma-poor)",
            "Neuroblastoma (Schwannian Stroma-Poor)",
            "NEUROBLASTOMA, MALIGNANT",
            "Neuroblastoma, NOS",
            "neuroblastoma",
        ],
        xrefs=[],
        associated_with=["umls:C2751421", "icdo:9500/3"],
        src_name="NCIt",
    )


@pytest.fixture(scope="module")
def nsclc():
    """Build fixture for non-small cell lung carcinoma"""
    return Disease(
        concept_id="ncit:C2926",
        label="Lung Non-Small Cell Carcinoma",
        aliases=[
            "Non Small Cell Lung Cancer NOS",
            "Non-Small Cell Lung Cancer",
            "Non-Small Cell Cancer of the Lung",
            "NSCLC",
            "non-small cell lung cancer",
            "Non-Small Cell Carcinoma of the Lung",
            "Non-Small Cell Cancer of Lung",
            "Non-small cell lung cancer, NOS",
            "Non-Small Cell Carcinoma of Lung",
            "NSCLC - Non-Small Cell Lung Cancer",
            "Non-Small Cell Lung Carcinoma",
        ],
        xrefs=[],
        associated_with=["umls:C0007131"],
    )


def test_concept_id_match(ncit, neuroblastoma, nsclc, compare_response):
    """Test that concept ID search resolves to correct record"""
    response = ncit.search("ncit:C3270")
    compare_response(response, MatchType.CONCEPT_ID, neuroblastoma)

    response = ncit.search("ncit:c2926")
    compare_response(response, MatchType.CONCEPT_ID, nsclc)

    response = ncit.search("NCIT:C2926")
    compare_response(response, MatchType.CONCEPT_ID, nsclc)

    response = ncit.search("C3270")
    compare_response(response, MatchType.CONCEPT_ID, neuroblastoma)

    response = ncit.search("3270")
    assert response.match_type == MatchType.NO_MATCH


def test_label_match(ncit, neuroblastoma, nsclc, compare_response):
    """Test that label search resolves to correct record."""
    response = ncit.search("Neuroblastoma")
    compare_response(response, MatchType.LABEL, neuroblastoma)

    response = ncit.search("NEUROBLASTOMA")
    compare_response(response, MatchType.LABEL, neuroblastoma)

    response = ncit.search("lung non-small cell carcinoma")
    compare_response(response, MatchType.LABEL, nsclc)

    response = ncit.search("lung non small cell carcinoma")
    assert response.match_type == MatchType.NO_MATCH


def test_alias_match(ncit, neuroblastoma, nsclc, compare_response):
    """Test that alias search resolves to correct record."""
    response = ncit.search("neuroblastoma, nos")
    compare_response(response, MatchType.ALIAS, neuroblastoma)

    response = ncit.search("neuroblastoma (Schwannian Stroma-Poor)")
    compare_response(response, MatchType.ALIAS, neuroblastoma)

    response = ncit.search("Neuroblastoma, Malignant")
    compare_response(response, MatchType.ALIAS, neuroblastoma)

    response = ncit.search("Neural Crest Tumor, Malignant")
    compare_response(response, MatchType.ALIAS, neuroblastoma)

    response = ncit.search("nsclc")
    compare_response(response, MatchType.ALIAS, nsclc)

    response = ncit.search("NSCLC - Non-Small Cell Lung Cancer")
    compare_response(response, MatchType.ALIAS, nsclc)

    response = ncit.search("neuroblastoma nbl")
    assert response.match_type == MatchType.NO_MATCH


def test_associated_with_match(ncit, neuroblastoma, nsclc, compare_response):
    """Test that associated_with search resolves to correct record."""
    response = ncit.search("icdo:9500/3")
    compare_response(response, MatchType.ASSOCIATED_WITH, neuroblastoma)

    response = ncit.search("umls:C0007131")
    compare_response(response, MatchType.ASSOCIATED_WITH, nsclc)


def test_no_animals(ncit):
    response = ncit.search("Rat Neuroblastoma")
    assert response.match_type == MatchType.NO_MATCH, "Some rats snuck in!"


def test_meta(ncit):
    """Test that meta field is correct."""
    response = ncit.search("neuroblastoma")
    assert response.source_meta_.data_license == "CC BY 4.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by/4.0/legalcode"
    )
    assert re.match(r"\d{2}\.\d{2}[a-z]", response.source_meta_.version)
    assert (
        response.source_meta_.data_url == "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/"
    )
    assert response.source_meta_.rdp_url == "http://reusabledata.org/ncit.html"
    assert not response.source_meta_.data_license_attributes.non_commercial
    assert not response.source_meta_.data_license_attributes.share_alike
    assert response.source_meta_.data_license_attributes.attribution



"""Module to load and init namespace at package level."""

from .base import DiseaseNormalizerEtlError
from .do import DO
from .merge import Merge
from .mondo import Mondo
from .ncit import NCIt
from .omim import OMIM
from .oncotree import OncoTree

__all__ = [
    "DO",
    "OMIM",
    "DiseaseNormalizerEtlError",
    "Merge",
    "Mondo",
    "NCIt",
    "OncoTree",
]



"""Build OMIM test data."""

import csv
from pathlib import Path

from disease.database import create_db
from disease.etl import OMIM

TEST_IDS = [
    "309200",
    "613065",
    "247640",
]

omim = OMIM(create_db())
omim._extract_data()
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "omim"
outfile_path = TEST_DATA_DIR / omim._data_file.name
outfile_path.parent.mkdir(exist_ok=True)

with omim._data_file.open() as f:
    rows = list(csv.reader(f, delimiter="\t"))

write_rows = []

# get headers
write_rows.append(rows.pop(0))
write_rows.append(rows.pop(0))
write_rows.append(rows.pop(0))

write_rows.extend(row for row in rows if row[0].startswith("#") or row[1] in TEST_IDS)

with outfile_path.open("w") as f:
    writer = csv.writer(f, delimiter="\t")
    writer.writerows(write_rows)



"""Apply manual data restrictions and annotations to extracted records."""

import csv
from pathlib import Path

from disease.schemas import SourceName


class Rules:
    """Store manually-generated data rules for modifying extracted source data.

    Use to provide consistency in edge cases for computational normalization, and
    correct possible curation errors.

    Initialize within each source's ETL class. This entails a small amount of repeated
    work, and could be revisited if the rules CSV becomes truly large, but makes for
    cleaner code in the meantime.

    Currently used to delete specific parameters from listlike fields, but could be
    expanded to use wildcards (e.g. to prohibit a value from being used in any field)
    or to manually add custom values to a field.
    """

    def __init__(self, source_name: SourceName) -> None:
        """Initialize rules class.
        :param source_name: name of source to use, for filtering unneeded rules
        """
        rules_path = Path(__file__).parent / "rules.csv"
        self.rules: dict[str, list[tuple[str, str]]] = {}
        with rules_path.open() as rules_file:
            reader = csv.DictReader(rules_file, delimiter=",")
            for row in reader:
                if row["source"] == source_name:
                    concept_id = row["concept_id"]
                    if not self.rules.get(concept_id):
                        self.rules[concept_id] = [(row["field"], row["value"])]
                    else:
                        self.rules[concept_id].append((row["field"], row["value"]))

    def apply_rules_to_disease(self, disease: dict) -> dict:
        """Apply all rules to a disease. First find relevant rules, then call the
        apply method.

        :param disease: disease object from ETL base
        :return: processed disease object
        """
        relevant_rules = self.rules.get(disease["concept_id"], [])
        for rule in relevant_rules:
            disease = self._apply_rule_to_field(disease, rule[0], rule[1])
        return disease

    def _apply_rule_to_field(
        self, disease: dict, field: str, value: str | list | dict | int | float
    ) -> dict:
        """Given a (field, value) rule, apply it to the given disease object.

        :param disease: disease object ready to load to DB
        :param field: name of object property field to check
        :param value: value to remove from field, if possible
        :return: disease object with rule applied
        """
        if field not in {"aliases", "xrefs", "associated_with"}:
            msg = "Non-scalar fields currently not implemented"
            raise ValueError(msg)
        field_data = set(disease.get(field, []))
        if value in field_data:
            field_data.remove(value)
            disease[field] = list(field_data)
        return disease



"""Main application for FastAPI"""

import html
import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from enum import Enum
from typing import Annotated

from fastapi import FastAPI, HTTPException, Query, Request

from disease import __version__
from disease.config import get_config
from disease.database.database import create_db
from disease.query import InvalidParameterException, QueryHandler
from disease.schemas import (
    APP_DESCRIPTION,
    LAB_EMAIL,
    LAB_WEBPAGE_URL,
    NormalizationService,
    SearchService,
    ServiceInfo,
    ServiceOrganization,
    ServiceType,
)
from disease.utils import initialize_logs


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator:
    """Perform operations that interact with the lifespan of the FastAPI instance.

    See https://fastapi.tiangolo.com/advanced/events/#lifespan.

    :param app: FastAPI instance
    """
    if get_config().debug:
        initialize_logs(logging.DEBUG)
    else:
        initialize_logs(logging.INFO)
    db = create_db()
    query_handler = QueryHandler(db)
    app.state.query_handler = query_handler
    yield
    db.close_connection()


app = FastAPI(
    title="VICC Disease Normalizer",
    description=APP_DESCRIPTION,
    version=__version__,
    contact={
        "name": "Alex H. Wagner",
        "email": LAB_EMAIL,
        "url": LAB_WEBPAGE_URL,
    },
    license={
        "name": "MIT",
        "url": "https://github.com/cancervariants/disease_normalization/blob/main/LICENSE",
    },
    docs_url="/disease",
    openapi_url="/disease/openapi.json",
    swagger_ui_parameters={"tryItOutEnabled": True},
    lifespan=lifespan,
)


class _Tag(str, Enum):
    """Define tag names for endpoints."""

    SEARCH = "Search"
    NORMALIZE = "Normalize"
    META = "Meta"


# endpoint description text
get_matches_summary = "Given query, provide highest matches from each source."
search_descr = "Search for disease term."
response_descr = "A response to a validly-formed query."
q_descr = "Disease term to search for."
incl_descr = (
    "Comma-separated list of source names to include in "
    "response. Will exclude all other sources. Will return HTTP "
    "status code 422: Unprocessable Entity if both 'incl' and "
    "'excl' parameters are given."
)
excl_descr = (
    "Comma-separated list of source names to exclude in "
    "response. Will include all other sources. Will return HTTP "
    "status code 422: Unprocessable Entity if both 'incl' and"
    "'excl' parameters are given."
)
normalize_description = (
    "Return merged strongest-match concept for query string provided by user."
)


@app.get(
    "/disease/search",
    summary=get_matches_summary,
    description=search_descr,
    operation_id="getQueryResponse",
    response_description=response_descr,
    response_model_exclude_none=True,
    tags=[_Tag.SEARCH],
)
def search(
    request: Request,
    q: Annotated[str, Query(description=q_descr)],
    incl: Annotated[str | None, Query(description=incl_descr)] = "",
    excl: Annotated[str | None, Query(description=excl_descr)] = "",
) -> SearchService:
    """For each source, return strongest-match concepts for query string
    provided by user.

    :param q: disease search term
    :param incl: sources to include
    :param excl: sources to excl
    :return: search results
    """
    query_handler = request.app.state.query_handler
    try:
        response = query_handler.search(html.unescape(q), incl=incl, excl=excl)
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response


merged_matches_summary = "Given query, provide normalized record."
merged_response_descr = "A response to a validly-formed query."
merged_q_descr = "Disease to normalize."


@app.get(
    "/disease/normalize",
    summary=merged_matches_summary,
    operation_id="getQuerymergedResponse",
    response_description=merged_response_descr,
    description=normalize_description,
    response_model_exclude_none=True,
    tags=[_Tag.NORMALIZE],
)
def normalize(
    request: Request,
    q: Annotated[str, Query(description=merged_q_descr)],
) -> NormalizationService:
    """Return strongest-match normalized concept for query string provided by
    user.

    :param q: disease search term
    """
    query_handler = request.app.state.query_handler
    try:
        response = query_handler.normalize(q)
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response


@app.get(
    "/disease/service-info",
    summary="Get basic service information",
    description="Retrieve service metadata, such as versioning and contact info. Structured in conformance with the [GA4GH service info API specification](https://www.ga4gh.org/product/service-info/)",
    tags=[_Tag.META],
)
def service_info() -> ServiceInfo:
    """Provide service info per GA4GH Service Info spec
    :return: conformant service info description
    """
    return ServiceInfo(
        organization=ServiceOrganization(),
        type=ServiceType(),
        environment=get_config().env,
    )



"""Test database classes."""

import os

import pytest

from disease.schemas import RecordType

IS_DDB = not os.environ.get("DISEASE_NORM_DB_URL", "").lower().startswith("postgres")
IS_TEST_ENV = os.environ.get("DISEASE_NORM_TEST", "").lower() == "true"


def test_tables_created(database):
    """Check that required tables are created."""
    existing_tables = database.list_tables()
    if database.__class__.__name__ == "PostgresDatabase":
        assert set(existing_tables) == {
            "disease_associations",
            "disease_labels",
            "disease_aliases",
            "disease_xrefs",
            "disease_concepts",
            "disease_merged",
            "disease_sources",
        }
    else:
        assert database.disease_table in existing_tables


@pytest.mark.skipif(not IS_DDB, reason="only applies to DynamoDB in test env")
def test_item_type(database):
    """Check that objects are tagged with item_type attribute."""
    from boto3.dynamodb.conditions import Key  # noqa: PLC0415

    filter_exp = Key("label_and_type").eq("ncit:c2926##identity")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "identity"

    filter_exp = Key("label_and_type").eq("neuroblastoma##label")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "label"

    filter_exp = Key("label_and_type").eq("umls:c0027819##associated_with")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "associated_with"

    filter_exp = Key("label_and_type").eq("oncotree:nbl##xref")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "xref"

    filter_exp = Key("label_and_type").eq("childhood liposarcoma##alias")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "alias"

    filter_exp = Key("label_and_type").eq("ncit:c9012##merger")
    item = database.diseases.query(KeyConditionExpression=filter_exp)["Items"][0]
    assert "item_type" in item
    assert item["item_type"] == "merger"


@pytest.mark.skipif(not IS_TEST_ENV, reason="not in test environment")
def database(db_fixture):
    """Perform basic test of get_all_records method.

    It's probably overkill (and unmaintainable) to do exact checks against every
    record, but fairly easy to check against expected counts and ensure that nothing
    is getting sent twice.
    """
    source_records = list(db_fixture.get_all_records(RecordType.IDENTITY))
    assert len(source_records) == 1463
    source_ids = {r["concept_id"] for r in source_records}
    assert len(source_ids) == 1463

    normalized_records = list(db_fixture.get_all_records(RecordType.MERGER))
    assert len(normalized_records) == 1391
    normalized_ids = {r["concept_id"] for r in normalized_records}
    assert len(normalized_ids) == 1391



"""Provide core database classes and helper methods.

Users shouldn't need to interact with the base class directly, but
:py:meth:`~disease.database.database.create_db` is the recommended way to create a
database connection.
"""

import abc
import sys
from collections.abc import Generator
from enum import Enum
from os import environ
from pathlib import Path
from typing import Any

import click

from disease.schemas import (
    RecordType,
    RefType,
    ServiceEnvironment,
    SourceMeta,
    SourceName,
)


class DatabaseException(Exception):  # noqa: N818
    """Create custom class for handling database exceptions"""


class DatabaseInitializationException(DatabaseException):
    """Create custom exception for errors during DB connection initialization."""


class DatabaseReadException(DatabaseException):
    """Create custom exception for lookup/read errors"""


class DatabaseWriteException(DatabaseException):
    """Create custom exception for write errors"""


class AbstractDatabase(abc.ABC):
    """Define a database interface."""

    @abc.abstractmethod
    def __init__(self, db_url: str | None = None, **db_args) -> None:
        """Initialize database instance.

        Generally, implementing classes should be able to construct a connection by
        something like a libpq URL. Any additional arguments or DB-specific parameters
        can be passed as keywords.

        :param db_url: address/connection description for database
        :param db_args: any DB implementation-specific parameters
        :raise DatabaseInitializationException: if initial setup fails
        """

    @abc.abstractmethod
    def list_tables(self) -> list[str]:
        """Return names of tables in database.

        :return: Table names in database
        """

    @staticmethod
    def _check_delete_okay() -> bool:
        """Check that environmental conditions permit DB deletion, and require
        confirmation.

        :raise DatabaseWriteException: if skip confirmation variable is set -- manual
        approval is required.
        """
        if environ.get(AWS_ENV_VAR_NAME, "") == AwsEnvName.PRODUCTION:
            if environ.get(SKIP_AWS_DB_ENV_NAME, "") == "true":
                msg = f"Must unset {SKIP_AWS_DB_ENV_NAME} env variable to enable drop_db()"
                raise DatabaseWriteException(msg)
            return click.confirm("Are you sure you want to delete existing data?")
        return True

    @abc.abstractmethod
    def drop_db(self) -> None:
        """Initiate total teardown of DB. Useful for quickly resetting the entirety of
        the data. Requires manual confirmation.

        :raise DatabaseWriteException: if called in a protected setting with
            confirmation silenced.
        """

    @abc.abstractmethod
    def check_schema_initialized(self) -> bool:
        """Check if database schema is properly initialized.

        :return: True if DB appears to be fully initialized, False otherwise
        """

    @abc.abstractmethod
    def check_tables_populated(self) -> bool:
        """Perform rudimentary checks to see if tables are populated.
        Emphasis is on rudimentary -- if some fiendish element has deleted half of the
        disease aliases, this method won't pick it up. It just wants to see if a few
        critical tables have at least a small number of records.

        :return: True if queries successful, false if DB appears empty
        """

    @abc.abstractmethod
    def initialize_db(self) -> None:
        """Perform all necessary parts of database setup. Should be tolerant of
        existing content -- ie, this method is also responsible for checking whether
        the DB is already set up.

        :raise DatabaseInitializationException: if initialization fails
        """

    @abc.abstractmethod
    def get_source_metadata(self, src_name: str | SourceName) -> SourceMeta | None:
        """Get license, versioning, data lookup, etc information for a source.

        :param src_name: name of the source to get data for
        :return: source metadata, if lookup is successful
        """

    @abc.abstractmethod
    def get_record_by_id(
        self, concept_id: str, case_sensitive: bool = True, merge: bool = False
    ) -> dict | None:
        """Fetch record corresponding to provided concept ID

        :param concept_id: concept ID for record
        :param case_sensitive: if true, performs exact lookup, which may be quicker.
            Otherwise, performs filter operation, which doesn't require correct casing.
        :param merge: if true, look for merged record; look for identity
            record otherwise.
        :return: complete record, if match is found; None otherwise
        """

    @abc.abstractmethod
    def get_refs_by_type(self, search_term: str, ref_type: RefType) -> list[str]:
        """Retrieve concept IDs for records matching the user's query. Other methods
        are responsible for actually retrieving full records.

        :param search_term: string to match against
        :param ref_type: type of match to look for.
        :return: list of associated concept IDs. Empty if lookup fails.
        """

    @abc.abstractmethod
    def get_all_concept_ids(self, source: SourceName | None = None) -> set[str]:
        """Retrieve all available concept IDs for use in generating normalized records.

        :param source: optionally, just get all IDs for a specific source
        :return: Set of concept IDs as strings.
        """

    @abc.abstractmethod
    def get_all_records(self, record_type: RecordType) -> Generator[dict, None, None]:
        """Retrieve all source or normalized records. Either return all source records,
        or all records that qualify as "normalized" (i.e., merged groups + source
        records that are otherwise ungrouped).
        For example,

        .. code-block:: pycon

           >>> from disease.database import create_db
           >>> from disease.schemas import RecordType
           >>> db = create_db()
           >>> for record in db.get_all_records(RecordType.MERGER):
           >>>     pass  # do something

        :param record_type: type of result to return
        :return: Generator that lazily provides records as they are retrieved
        """

    @abc.abstractmethod
    def add_source_metadata(self, src_name: SourceName, meta: SourceMeta) -> None:
        """Add new source metadata entry.

        :param src_name: name of source
        :param meta: known source attributes
        :raise DatabaseWriteException: if write fails
        """

    @abc.abstractmethod
    def add_record(self, record: dict, src_name: SourceName) -> None:
        """Add new record to database.

        :param record: record to upload
        :param src_name: name of source for record.
        """

    @abc.abstractmethod
    def add_merged_record(self, record: dict) -> None:
        """Add merged record to database.

        :param record: merged record to add
        """

    @abc.abstractmethod
    def update_merge_ref(self, concept_id: str, merge_ref: Any) -> None:  # noqa: ANN401
        """Update the merged record reference of an individual record to a new value.

        :param concept_id: record to update
        :param merge_ref: new ref value
        :raise DatabaseWriteException: if attempting to update non-existent record
        """

    @abc.abstractmethod
    def delete_normalized_concepts(self) -> None:
        """Remove merged records from the database. Use when performing a new update
        of normalized data.

        :raise DatabaseReadException: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteException: if deletion call fails
        """

    @abc.abstractmethod
    def delete_source(self, src_name: SourceName) -> None:
        """Delete all data for a source. Use when updating source data.

        :param src_name: name of source to delete
        :raise DatabaseReadException: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteException: if deletion call fails
        """

    @abc.abstractmethod
    def complete_write_transaction(self) -> None:
        """Conclude transaction or batch writing if relevant."""

    @abc.abstractmethod
    def close_connection(self) -> None:
        """Perform any manual connection closure procedures if necessary."""

    @abc.abstractmethod
    def load_from_remote(self, url: str | None = None) -> None:
        """Load DB from remote dump. Warning: Deletes all existing data.

        :param url: remote location to retrieve gzipped dump file from
        :raise: NotImplementedError if not supported by DB
        """

    @abc.abstractmethod
    def export_db(self, export_location: Path) -> None:
        """Dump DB to specified location.

        :param export_location: path to save DB dump at
        :raise: NotImplementedError if not supported by DB
        """


# can be set to either `Dev`, `Staging`, or `Prod`
# ONLY set when wanting to access aws instance
AWS_ENV_VAR_NAME = "DISEASE_NORM_ENV"

# Set to "true" if want to skip db confirmation check. Should ONLY be used for
# deployment needs
SKIP_AWS_DB_ENV_NAME = "SKIP_AWS_CONFIRMATION"


class AwsEnvName(str, Enum):
    """AWS environment name that is being used"""

    DEVELOPMENT = ServiceEnvironment.DEV.value
    STAGING = ServiceEnvironment.STAGING.value
    PRODUCTION = ServiceEnvironment.PROD.value


VALID_AWS_ENV_NAMES = {v.value for v in AwsEnvName.__members__.values()}


def confirm_aws_db_use(env_name: str) -> None:
    """Check to ensure that AWS instance should actually be used."""
    if click.confirm(
        f"Are you sure you want to use the AWS {env_name} database?", default=False
    ):
        click.echo(f"***DISEASE AWS {env_name.upper()} DATABASE IN USE***")
    else:
        click.echo("Exiting.")
        sys.exit()


def create_db(
    db_url: str | None = None, aws_instance: bool = False
) -> AbstractDatabase:
    """Database factory method. Checks environment variables and provided parameters
    and creates a DB instance.

    Generally prefers to return a DynamoDB instance, unless all DDB-relevant
    environment variables are unset and a libpq-compliant URI is passed to
    ``db_url``.

    :param db_url: address to database instance
    :param aws_instance: use hosted DynamoDB instance, not local DB
    :return: constructed Database instance
    """
    aws_env_var_set = AWS_ENV_VAR_NAME in environ
    if aws_env_var_set or aws_instance:
        from disease.database.dynamodb import DynamoDbDatabase  # noqa: PLC0415

        db = DynamoDbDatabase()
    else:
        if db_url:
            endpoint_url = db_url
        elif "DISEASE_NORM_DB_URL" in environ:
            endpoint_url = environ["DISEASE_NORM_DB_URL"]
        else:
            endpoint_url = "http://localhost:8000"

        # prefer DynamoDB unless connection explicitly reads like a libpq URI
        if endpoint_url.startswith("postgres"):
            from disease.database.postgresql import PostgresDatabase  # noqa: PLC0415

            db = PostgresDatabase(endpoint_url)
        else:
            from disease.database.dynamodb import DynamoDbDatabase  # noqa: PLC0415

            db = DynamoDbDatabase(endpoint_url)
    return db



"""Get Human Disease Ontology data."""

import owlready2 as owl
from tqdm import tqdm

from disease import PREFIX_LOOKUP
from disease.etl.base import OWLBase
from disease.schemas import NamespacePrefix, SourceMeta

DO_PREFIX_LOOKUP = {
    "EFO": NamespacePrefix.EFO.value,
    "GARD": NamespacePrefix.GARD.value,
    "ICDO": NamespacePrefix.ICDO.value,
    "MESH": NamespacePrefix.MESH.value,
    "NCI": NamespacePrefix.NCIT.value,
    "ORDO": NamespacePrefix.ORPHANET.value,
    "UMLS_CUI": NamespacePrefix.UMLS.value,
    "ICD9CM": NamespacePrefix.ICD9CM.value,
    "ICD10CM": NamespacePrefix.ICD10CM.value,
    "MIM": NamespacePrefix.OMIM.value,
    "KEGG": NamespacePrefix.KEGG.value,
    "MEDDRA": NamespacePrefix.MEDDRA.value,
}


class DO(OWLBase):
    """Disease Ontology ETL class."""

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC0 1.0",
            data_license_url="https://creativecommons.org/publicdomain/zero/1.0/legalcode",
            version=self._version,
            data_url="http://www.obofoundry.org/ontology/doid.html",
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            },
        )
        self._database.add_source_metadata(self._src_name, metadata)

    def _transform_data(self) -> None:
        """Transform source data and send to loading method."""
        do = owl.get_ontology(self._data_file.absolute().as_uri()).load()
        disease_uri = "http://purl.obolibrary.org/obo/DOID_4"
        diseases = self._get_subclasses(
            disease_uri, owl.default_world.as_rdflib_graph()
        )
        # use "disease of cellular proliferation" to capture related non- or pre-malignant
        # disease terms
        cancers_uri = "http://purl.obolibrary.org/obo/DOID_14566"
        cancers = self._get_subclasses(cancers_uri, owl.default_world.as_rdflib_graph())
        for uri in tqdm(diseases, ncols=80, disable=self._silent):
            disease_class = do.search(iri=uri)[0]
            if disease_class.deprecated:
                continue

            concept_id = f"{NamespacePrefix.DO.value}:{uri.split('_')[-1]}"
            label = disease_class.label[0]

            synonyms = disease_class.hasExactSynonym
            aliases = list({s for s in synonyms if s != label}) if synonyms else []

            xrefs = []
            associated_with = []
            db_associated_with = set(disease_class.hasDbXref)
            for xref in db_associated_with:
                prefix, id_no = xref.split(":", 1)
                normed_prefix = DO_PREFIX_LOOKUP.get(prefix)
                if normed_prefix:
                    xref_no = f"{normed_prefix}:{id_no}"
                    if normed_prefix.lower() in PREFIX_LOOKUP:
                        if normed_prefix == NamespacePrefix.OMIM and id_no.startswith(
                            "PS"
                        ):
                            associated_with.append(xref_no)
                        else:
                            xrefs.append(xref_no)
                    else:
                        associated_with.append(xref_no)

            disease = {
                "concept_id": concept_id,
                "label": label,
                "aliases": aliases,
                "xrefs": xrefs,
                "associated_with": associated_with,
            }

            if uri in cancers:
                disease["oncologic_disease"] = True
            self._load_disease(disease)



"""Test merged record generation."""

import os
from collections.abc import Callable

import pytest

from disease import SOURCES_FOR_MERGE
from disease.database import AWS_ENV_VAR_NAME
from disease.database.database import create_db
from disease.etl import DO, OMIM, Merge, Mondo, NCIt, OncoTree
from disease.schemas import SourceName


@pytest.fixture(scope="module")
def merge_instance(test_source: Callable, is_test_env: bool):
    """Provide fixture for ETL merge class.

    If in a test environment (e.g. CI) this method will attempt to load any missing
    source data, and then perform merged record generation.
    """
    database = create_db()
    if is_test_env:
        if os.environ.get(AWS_ENV_VAR_NAME):
            msg = f"Running the full disease ETL pipeline test on an AWS environment is forbidden -- either unset {AWS_ENV_VAR_NAME} or unset DISEASE_NORM_TEST"
            raise AssertionError(msg)
        for SourceClass in (Mondo, DO, NCIt, OncoTree, OMIM):  # noqa: N806
            if not database.get_source_metadata(SourceName(SourceClass.__name__)):
                test_source(SourceClass)

    m = Merge(database)
    if is_test_env:
        concept_ids = set()
        for source in SOURCES_FOR_MERGE:
            concept_ids |= database.get_all_concept_ids(source)
        m.create_merged_concepts(concept_ids)
    return m


@pytest.fixture(scope="module")
def neuroblastoma():
    """Create neuroblastoma fixture."""
    return {
        "concept_id": "ncit:C3270",
        "item_type": "merger",
        "xrefs": ["mondo:0005072", "oncotree:NBL", "DOID:769"],
        "label": "Neuroblastoma",
        "aliases": [
            "NB",
            "neuroblastoma",
            "Neural Crest Tumor, Malignant",
            "Neuroblastoma (Schwannian Stroma-poor)",
            "neuroblastoma (Schwannian Stroma-poor)",
            "Neuroblastoma (Schwannian Stroma-Poor)",
            "Neuroblastoma, NOS",
            "NEUROBLASTOMA, MALIGNANT",
            "Neuroblastoma (NBL)",
            "neural Crest tumor, malignant",
            "neuroblastoma, malignant",
        ],
        "associated_with": [
            "efo:0000621",
            "gard:7185",
            "icdo:9500/3",
            "mesh:D009447",
            "medgen:18012",
            "orphanet:635",
            "umls:C0027819",
            "umls:C2751421",
        ],
        "oncologic_disease": True,
    }


@pytest.fixture(scope="module")
def lnscc():
    """Create lung non small cell carcinoma fixture"""
    return {
        "concept_id": "ncit:C2926",
        "xrefs": ["mondo:0005233", "oncotree:NSCLC", "DOID:3908"],
        "label": "Lung Non-Small Cell Carcinoma",
        "aliases": [
            "NSCLC - Non-Small Cell Lung Cancer",
            "NSCLC - non-small cell lung cancer",
            "NSCLC",
            "Non Small Cell Lung Cancer NOS",
            "Non-Small Cell Cancer of Lung",
            "Non-Small Cell Cancer of the Lung",
            "Non-Small Cell Carcinoma of Lung",
            "Non-Small Cell Carcinoma of the Lung",
            "Non-Small Cell Lung Cancer",
            "Non-Small Cell Lung Carcinoma",
            "Non-small cell lung cancer",
            "Non-small cell lung cancer, NOS",
            "non-small cell cancer of lung",
            "non-small cell cancer of the lung",
            "non-small cell carcinoma of lung",
            "non-small cell carcinoma of the lung",
            "non-small cell lung cancer",
            "non-small cell lung carcinoma (disease)",
            "non-small cell lung carcinoma",
        ],
        "associated_with": [
            "umls:C0007131",
            "mesh:D002289",
            "efo:0003060",
            "kegg.disease:05223",
            "medgen:40104",
        ],
        "item_type": "merger",
        "oncologic_disease": True,
    }


@pytest.fixture(scope="module")
def richter():
    """Create Richter Syndrome fixture"""
    return {
        "concept_id": "ncit:C35424",
        "xrefs": ["mondo:0002083", "DOID:1703"],
        "label": "Richter Syndrome",
        "aliases": [
            "Richter's Syndrome",
            "Richter syndrome",
            "Richter transformation",
            "Richter's Transformation",
            "Richter's syndrome",
            "Richter's transformation",
        ],
        "associated_with": [
            "umls:C0349631",
            "gard:7578",
            "icd10.cm:C91.1",
            "medgen:91159",
        ],
        "item_type": "merger",
        "oncologic_disease": True,
    }


@pytest.fixture(scope="module")
def ped_liposarcoma():
    """Create pediatric liposarcoma fixture."""
    return {
        "concept_id": "ncit:C8091",
        "xrefs": ["mondo:0003587", "DOID:5695"],
        "label": "Childhood Liposarcoma",
        "aliases": [
            "liposarcoma",
            "Liposarcoma",
            "Pediatric Liposarcoma",
            "childhood liposarcoma",
            "pediatric liposarcoma",
        ],
        "associated_with": ["umls:C0279984", "medgen:83580"],
        "pediatric_disease": True,
        "oncologic_disease": True,
        "item_type": "merger",
    }


@pytest.fixture(scope="module")
def teratoma():
    """Create fixture for adult cystic teratoma."""
    return {
        "concept_id": "ncit:C9012",
        "xrefs": ["mondo:0004099", "DOID:7079"],
        "label": "Adult Cystic Teratoma",
        "aliases": [
            "Adult cystic teratoma",
            "cystic teratoma of adults",
            "adult cystic teratoma",
        ],
        "associated_with": ["icdo:9080/0", "umls:C1368888", "medgen:235084"],
        "item_type": "merger",
        "oncologic_disease": True,
    }


@pytest.fixture(scope="module")
def mafd2():
    """Create a fixture for major affective disorder 2. Tests whether a
    deprecated DO reference is filtered out.
    """
    return {
        "item_type": "merger",
        "concept_id": "mondo:0010648",
        "label": "major affective disorder 2",
        "aliases": [
            "BIPOLAR AFFECTIVE DISORDER",
            "BPAD",
            "MAFD2",
            "MANIC-DEPRESSIVE ILLNESS",
            "MANIC-DEPRESSIVE PSYCHOSIS, X-LINKED",
            "MDI",
            "MDX",
            "major affective disorder 2, X-linked dominant",
            "MAJOR affective disorder 2",
            "bipolar affective disorder",
            "major affective disorder 2",
            "manic-depressive illness",
            "manic-depressive psychosis, X-linked",
        ],
        "xrefs": [
            "MIM:309200",
        ],
        "associated_with": ["mesh:C564108", "medgen:326975", "umls:C1839839"],
    }


@pytest.fixture(scope="module")
def record_id_groups():
    """Fixture for concept ID group input."""
    return {
        "neuroblastoma": ["ncit:C3270", "mondo:0005072", "DOID:769", "oncotree:NBL"],
        "lnscc": ["ncit:C2926", "mondo:0005233", "DOID:3908", "oncotree:NSCLC"],
        "richter": ["ncit:C35424", "mondo:0002083", "DOID:1703"],
        "ped_liposarcoma": ["ncit:C8091", "mondo:0003587", "DOID:5695"],
        "teratoma": ["ncit:C9012", "mondo:0004099", "DOID:7079"],
        "mafd2": ["mondo:0010648", "MIM:309200"],
    }


def compare_merged_records(actual, fixture):
    """Verify correctness of merged DB record."""
    assert actual["concept_id"] == fixture["concept_id"]
    assert ("xrefs" in actual) == ("xrefs" in fixture)
    if "xrefs" in actual:
        assert set(actual["xrefs"]) == set(fixture["xrefs"])

    assert ("label" in actual) == ("label" in fixture)
    if "label" in actual or "label" in fixture:
        assert actual["label"] == fixture["label"]

    assert ("aliases" in actual) == ("aliases" in fixture)
    if "aliases" in actual or "aliases" in fixture:
        assert set(actual["aliases"]) == set(fixture["aliases"])

    assert ("associated_with" in actual) == ("associated_with" in fixture)
    if "associated_with" in actual or "associated_with" in fixture:
        assert set(actual["associated_with"]) == set(fixture["associated_with"])

    assert ("pediatric_disease" in actual) == ("pediatric_disease" in fixture)
    if "pediatric_disease" in actual or "pediatric_disease" in fixture:
        assert actual["pediatric_disease"] == fixture["pediatric_disease"]

    assert ("oncologic_disease" in actual) == ("oncologic_disease" in fixture)
    if "oncologic_disease" in actual or "oncologic_disease" in fixture:
        assert actual["oncologic_disease"] == fixture["oncologic_disease"]


def test_generate_merged_record(
    merge_instance,
    record_id_groups,
    neuroblastoma,
    lnscc,
    richter,
    ped_liposarcoma,
    teratoma,
    mafd2,
):
    """Test generation of individual merged record."""
    neuroblastoma_ids = record_id_groups["neuroblastoma"]
    response, r_ids = merge_instance._generate_merged_record(neuroblastoma_ids)
    assert set(r_ids) == set(neuroblastoma_ids)
    compare_merged_records(response, neuroblastoma)

    lnscc_ids = record_id_groups["lnscc"]
    response, r_ids = merge_instance._generate_merged_record(lnscc_ids)
    assert set(r_ids) == set(lnscc_ids)
    compare_merged_records(response, lnscc)

    richter_ids = record_id_groups["richter"]
    response, r_ids = merge_instance._generate_merged_record(richter_ids)
    assert set(r_ids) == set(richter_ids)
    compare_merged_records(response, richter)

    ped_liposarcoma_ids = record_id_groups["ped_liposarcoma"]
    response, r_ids = merge_instance._generate_merged_record(ped_liposarcoma_ids)
    assert set(r_ids) == set(ped_liposarcoma_ids)
    compare_merged_records(response, ped_liposarcoma)

    teratoma_ids = record_id_groups["teratoma"]
    response, r_ids = merge_instance._generate_merged_record(teratoma_ids)
    assert set(r_ids) == set(teratoma_ids)
    compare_merged_records(response, teratoma)

    mafd2_ids = record_id_groups["mafd2"]
    response, r_ids = merge_instance._generate_merged_record(mafd2_ids)
    assert set(r_ids) == {"mondo:0010648", "MIM:309200"}
    compare_merged_records(response, mafd2)



"""Module to test the schemas module."""

from disease.schemas import NAMESPACE_TO_SYSTEM_URI, NamespacePrefix


def test_namespace_to_system_uri():
    """Ensure that each NamespacePrefix is included in NAMESPACE_TO_SYSTEM_URI"""
    for v in NamespacePrefix.__members__.values():
        assert v in NAMESPACE_TO_SYSTEM_URI, v
        assert NAMESPACE_TO_SYSTEM_URI[v].startswith("http"), v



"""Test the emit_warnings function."""

from disease.query import QueryHandler


def test_emit_warnings(database):
    """Test that emit_warnings works correctly."""
    expected_warnings = {"nbsp": "Query contains non breaking space characters."}
    query_handler = QueryHandler(database)

    # Test emit no warnings
    actual_warnings = query_handler._emit_warnings("CISPLATIN")
    assert actual_warnings is None

    # Test emit warnings
    actual_warnings = query_handler._emit_warnings("CIS PLATIN")
    assert actual_warnings == actual_warnings

    actual_warnings = query_handler._emit_warnings("CIS\u00a0platin")
    assert expected_warnings == actual_warnings

    actual_warnings = query_handler._emit_warnings("CIS&nbsp;platin")
    assert expected_warnings == actual_warnings

    actual_warnings = query_handler._emit_warnings("CIS\xa0platin")
    assert expected_warnings == actual_warnings



"""Perform basic tests of endpoint branches.
We already have tests and data validation to ensure correctness of the underlying
response objects -- here, we're checking for bad branch logic and for basic assurances
that routes integrate correctly with query methods.
"""

from pathlib import Path

import jsonschema
import pytest
import pytest_asyncio
import yaml
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from disease.main import app


@pytest_asyncio.fixture
async def async_app():
    async with LifespanManager(app) as manager:
        yield manager.app


@pytest_asyncio.fixture
async def api_client(async_app):
    async with AsyncClient(
        transport=ASGITransport(async_app), base_url="http://tests"
    ) as client:
        yield client


@pytest.mark.asyncio
async def test_search(api_client: AsyncClient):
    """Test /search endpoint."""
    response = await api_client.get("/disease/search?q=neuroblastoma")
    assert response.status_code == 200
    assert (
        response.json()["source_matches"]["Mondo"]["records"][0]["concept_id"]
        == "mondo:0005072"
    )

    response = await api_client.get("/disease/search?q=neuroblastoma&incl=sdkl")
    assert response.status_code == 422

    response = await api_client.get("/disease/search")
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_normalize(api_client: AsyncClient):
    """Test /normalize endpoint."""
    response = await api_client.get("/disease/normalize?q=neuroblastoma")
    assert response.status_code == 200
    assert response.json()["disease"]["primaryCoding"] == {
        "id": "ncit:C3270",
        "code": "C3270",
        "system": "https://ncit.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code=",
    }

    response = await api_client.get("/disease/normalize")
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_service_info(api_client: AsyncClient, test_data_dir: Path):
    response = await api_client.get("/disease/service-info")
    response.raise_for_status()

    with (test_data_dir / "service_info_openapi.yaml").open() as f:
        spec = yaml.safe_load(f)

    resp_schema = spec["paths"]["/service-info"]["get"]["responses"]["200"]["content"][
        "application/json"
    ]["schema"]

    resolver = jsonschema.RefResolver.from_schema(spec)
    data = response.json()
    jsonschema.validate(instance=data, schema=resp_schema, resolver=resolver)
